"""This module only contains the Assembler class.

Author: Thomas Oberbichler
"""

class Assembler(object):
    """An Assembler helps to generate system matrices/vectors from elements.

    Attributes
    ----------
    dofs : list
        List of all dofs in the system. The Dofs with dirichlet constraints are at the end.
    dof_indices : dict
        Dictionary containing the index of each dof. The index is equal to the position of the dof
        in the dofs-list.
    dof_count : int
        Total number of dofs.
    free_dof_count : int
        Number of dofs without dirichlet constraints.
    fixed_dof_count : int
        Number of dofs with dirichlet constraints.
    element_freedom_table : list
        List with tuples containing the elements and the coresponding dof indices.
    """

    def __init__(self, model):
        """Create a new Assembler

        Parameters
        ----------
        model : Model
            Model to assemble.
        """

        # --- dof indices

        processed_dofs = set()

        free_dofs = list()
        fixed_dofs = list()

        for element in model.elements:
            for dof in element.dofs:
                if dof in processed_dofs:
                    continue
                else:
                    processed_dofs.add(dof)

                if not model.dof_is_active(dof):
                    fixed_dofs.append(dof)
                else:
                    free_dofs.append(dof)

        dofs = free_dofs + fixed_dofs

        dof_indices = {dof: index for index, dof in enumerate(dofs)}


        # --- element freedom table

        element_freedom_table = list()

        for element in model.elements:
            indices = [dof_indices[dof] for dof in element.dofs]

            element_freedom_table.append((element, indices))


        # --- store

        self.dofs = dofs
        self.dof_indices = dof_indices
        self.dof_count = len(dofs)
        self.free_dof_count = len(free_dofs)
        self.fixed_dof_count = len(fixed_dofs)
        self.element_freedom_table = element_freedom_table

    def index_of_dof(self, dof):
        """Get the index of the given dof.

        Parameters
        ----------
        dof : object
            Dof at the index.

        Returns
        -------
        index : int
            Index of the given dof.
        """
        return self.dof_indices[dof]

    def dof_at_index(self, index):
        """Get the dof at the given index.

        Parameters
        ----------
        index : int
            Index of the dof.

        Returns
        -------
        dof : object
            Dof at the given index.
        """
        return self.dofs[index]

    @property
    def free_dofs(self):
        """Get a list with the unconstrained dofs

        Returns
        -------
        free_dofs : list
            List with the unconstrained dofs
        """
        return self.dofs[:self.free_dof_count]

    def assemble_matrix(self, system_matrix, calculate_element_matrix):
        """Assemble element matrices into a system matrix.

        Parameters
        ----------
        system_matrix : ndarray
            System matrix to store the results. The results are added to the existing values.
        calculate_element_matrix : function Element -> ndarray
            Function to calculate the element matrix.
        """
        for element, indices in self.element_freedom_table:
            element_matrix = calculate_element_matrix(element)

            if element_matrix is None:
                continue

            for element_row, system_row in enumerate(indices):
                for element_col, system_col in enumerate(indices):
                    value = element_matrix[element_row, element_col]
                    system_matrix[system_row, system_col] += value

    def assemble_vector(self, system_vector, calculate_element_vector):
        """Assemble element vectors into a system vector.

        Parameters
        ----------
        system_vector : ndarray
            System vector to store the results. The results are added to the existing values.
        calculate_element_vector : function Element -> ndarray
            Function to calculate the element vector.
        """
        for element, indices in self.element_freedom_table:
            element_vector = calculate_element_vector(element)

            if element_vector is None:
                continue

            for element_row, system_row in enumerate(indices):
                system_vector[system_row] += element_vector[element_row]



"""
NFEM teaching tool

A module for the non linear static analysis of 3D truss problems.
A light weight object oriented FEM code and some usefull postprocessing tools.
"""

__version__ = '1.9.8'

from .model import Model
from .node import Node
from .truss import Truss
from .assembler import Assembler

from .newton_raphson import newton_raphson_solve
from .path_following_method import LoadControl
from .path_following_method import DisplacementControl
from .path_following_method import ArcLengthControl

from .bracketing import bracketing

from .visualization import *


def info():
    print(f'--------------------------------------------------------------------------------')
    print(f'                                                                                ')
    print(f'                      _   ________________  ___                                 ')
    print(f'                     / | / / ____/ ____/  |/  /                                 ')
    print(f'                    /  |/ / /_  / __/ / /|_/ /                                  ')
    print(f'                   / /|  / __/ / /___/ /  / /                                   ')
    print(f'                  /_/ |_/_/   /_____/_/  /_/  Teaching Tool                     ')
    print(f'                                                                                ')
    print(f'  Authors:   Armin Geiser, Aditya Ghantasala, Thomas Oberbichler, Klaus Sautter ')
    print(f'             Mahmoud Zidan                                                      ')
    print(f'  Copyright: © 2018-2020 TUM Statik                                             ')
    print(f'  Version:   {__version__}                                                      ')
    print(f'                                                                                ')
    print(f'  This is a teaching tool! All results without warranty.                        ')
    print(f'                                                                                ')
    print(f'--------------------------------------------------------------------------------')


__all__ = [
    'Model',
    'Node',
    'Truss',
    'Assembler',
    'newton_raphson_solve',
    'LoadControl',
    'DisplacementControl',
    'ArcLengthControl',
    'bracketing',
    'info',
    'show_load_displacement_curve',
    'show_animation',
    'show_deformation_plot',
    'Plot2D',
]

try:
    from .visualization import interact
    __all__ += [
        'interact',
    ]
except Exception as _:
    pass



"""This module contains path-following strategies for solving nonlinear systems.

Author: Armin Geiser
"""

import numpy as np

from .assembler import Assembler

class PathFollowingMethod(object):

    def scale_predictor(self, model):
        raise NotImplementedError

    def calculate_constraint(self, model):
        #returns c
        raise NotImplementedError

    def calculate_derivatives(self, model, dc):
        #returns dc_du, dc_dLambda
        raise NotImplementedError

class LoadControl(PathFollowingMethod):
    """The LoadControl adds a constraint to the non linear problem that ensures
        a prescribed load factor at the equilibrium point.

    Attributes
    ----------
    lam_hat : float
        Prescribed value for the load factor
    """

    def __init__(self, model):
        """Create a new LoadControl

        Parameters
        ----------
        lam_hat : float
            Prescribed value for the load factor
        """
        super(LoadControl, self).__init__()
        self.lam_hat = model.lam

    def calculate_constraint(self, model):
        """Calculates the constraint

        Parameters
        ----------
        model : Model
            Model for which the constraint is calculated.

        Returns
        ----------
        constraint : float
            Value of the constraint.
        """
        return model.lam - self.lam_hat

    def calculate_derivatives(self, model, dc):
        """Assembles the constraint derivative [dc/du, dc/dlam] into the vector dc

        Parameters
        ----------
        model : Model
            Model for which the constraint derivative is calculated.
        dc : ndarray
            System vector to store the results. Existing values are overwritten.
        """
        dc.fill(0.0)
        dc[-1] = 1.0

class DisplacementControl(PathFollowingMethod):
    """The DisplacementControl adds a constraint to the non linear problem that ensures
        a prescribed displacement of the controlled dof at the equilibrium point.

    Attributes
    ----------
    dof : Object
        Controlled dof
    displacement_hat : float
        Prescribed value for the controlled dof
    """
    def __init__(self, model, dof):
        """Create a new DisplacementControl

        Parameters
        ----------
        dof : Object
            Controlled dof
        displacement_hat : float
            Prescribed value for the controlled dof
        """
        super(DisplacementControl, self).__init__()
        self.displacement_hat = model.get_dof_state(dof)
        self.dof = dof

    def calculate_constraint(self, model):
        """Calculates the constraint

        Parameters
        ----------
        model : Model
            Model for which the constraint is calculated.

        Returns
        ----------
        constraint : float
            Value of the constraint.
        """
        dof = self.dof
        displacement_hat = self.displacement_hat

        displacement = model.get_dof_state(dof)

        return displacement - displacement_hat

    def calculate_derivatives(self, model, dc):
        """Assembles the constraint derivative [dc/du, dc/dlam] into the vector dc

        Parameters
        ----------
        model : Model
            Model for which the constraint derivative is calculated.
        dc : ndarray
            System vector to store the results. Existing values are overwritten.
        """
        dc.fill(0.0)
        assembler = Assembler(model)
        index = assembler.index_of_dof(self.dof)
        dc[index] = 1.0

class ArcLengthControl(PathFollowingMethod):
    """The ArcLengthControl adds a constraint to the non linear problem that ensures
        a prescribed length of the increment between the new and the last equilibrium point.

    Attributes
    ----------
    l_hat : float
        Prescribed value for the length of the increment (l2 norm)
    """
    def __init__(self, model):
        """Create a new ArcLengthControl

        Parameters
        ----------
        l_hat : float
            Prescribed value for the length of the increment (l2 norm)
        """
        super(ArcLengthControl, self).__init__()
        self.squared_l_hat = self._calculate_squared_predictor_length(model)

    def calculate_constraint(self, model):
        """Calculates the constraint

        Parameters
        ----------
        model : Model
            Model for which the constraint is calculated.

        Returns
        ----------
        constraint : float
            Value of the constraint.
        """
        squared_l_hat = self.squared_l_hat
        squared_l = self._calculate_squared_predictor_length(model)

        return squared_l - squared_l_hat

    def calculate_derivatives(self, model, dc):
        """Assembles the constraint derivative [dc/du, dc/dlam] into the vector dc

        Parameters
        ----------
        model : Model
            Model for which the constraint derivative is calculated.
        dc : ndarray
            System vector to store the results. Existing values are overwritten.
        """
        dc.fill(0.0)

        assembler = Assembler(model)
        previous_model = model.get_previous_model()

        for index, dof in enumerate(assembler.free_dofs):
            current_value = model.get_dof_state(dof)
            previous_value = previous_model.get_dof_state(dof)

            dc[index] = 2 * (current_value - previous_value)

        dc[-1] = 2 * (model.lam - model.get_previous_model().lam)

    def _calculate_squared_predictor_length(self, model):
        previous_model = model.get_previous_model()

        squared_l = 0.0

        for node, previous_node in zip(model.nodes, previous_model.nodes):
            dx, dy, dz = node.location - previous_node.location
            squared_l += dx**2 + dy**2 + dz**2

        delta_lam = model.lam - previous_model.lam

        squared_l += delta_lam**2

        return squared_l



"""This module contains helpers for visualize data.

Authors: Klaus Sautter, Thomas Oberbichler, Armin Geiser
"""

from IPython.display import HTML
import numpy as np
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from matplotlib.figure import Figure
from mpl_toolkits.mplot3d.art3d import Line3DCollection
import matplotlib.animation as anim

from .plot_symbols import get_force_arrow, get_tet4_polygons, get_dof_arrow

from ..truss import Truss
from ..single_load import SingleLoad
from ..model import ModelStatus


def get_bounding_box(models):
    nodes = [node for model in models for node in model.nodes]

    min_x = min(node.x for node in nodes)
    max_x = max(node.x for node in nodes)

    min_y = min(node.y for node in nodes)
    max_y = max(node.y for node in nodes)

    min_z = min(node.z for node in nodes)
    max_z = max(node.z for node in nodes)

    return min_x, max_x, min_y, max_y, min_z, max_z


def plot_scaled_model(ax, model, color, **options):
    scaling_factor = options.get('plot/scaling_factor', None)
    if not scaling_factor:
        # autoscaling max u to 10% of bounding_box
        bounding_box = get_bounding_box([model.get_initial_model()])
        min_x, max_x, min_y, max_y, min_z, max_z = bounding_box
        max_delta = max(max_x-min_x, max_y-min_y, max_z-min_z)
        max_u = max(abs(node.u) for node in model.nodes)
        max_v = max(abs(node.v) for node in model.nodes)
        max_w = max(abs(node.w) for node in model.nodes)
        max_def = max(max_u, max_v, max_w)
        scaling_factor = max_delta / max_def * 0.1

    lines = list()

    for element in model.elements:
        if type(element) == Truss:
            node_a = element.node_a
            node_b = element.node_b

            b = [node_b.reference_x+scaling_factor*node_b.u, node_b.reference_y+scaling_factor*node_b.v, node_b.reference_z+scaling_factor*node_b.w]
            a = [node_a.reference_x+scaling_factor*node_a.u, node_a.reference_y+scaling_factor*node_a.v, node_a.reference_z+scaling_factor*node_a.w]

            lines.append([a, b])

    lc = Line3DCollection(lines, colors=color, linewidths=2)

    ax.add_collection(lc)

    plot_symbols(ax, model, color, initial=False, **options)


def plot_model(ax, model, color, initial, **options):
    lines = list()

    for element in model.elements:
        if type(element) == Truss:
            node_a = element.node_a
            node_b = element.node_b

            a = [node_a.reference_x, node_a.reference_y, node_a.reference_z] if initial else [node_a.x, node_a.y, node_a.z]
            b = [node_b.reference_x, node_b.reference_y, node_b.reference_z] if initial else [node_b.x, node_b.y, node_b.z]

            lines.append([a, b])

    lc = Line3DCollection(lines, colors=color, linewidths=2)

    ax.add_collection(lc)

    plot_symbols(ax, model, color, initial, **options)


def plot_symbols(ax, model, color, initial, **options):
    if options.get('plot/dirichlet', True):
        plot_boundary_conditions(ax, model, initial, **options)
    if options.get('plot/neumann', True):
        plot_forces(ax, model, initial, **options)
    if options.get('plot/highlight_dof', False):
        plot_dof_higlight(ax, model, initial, **options)


def get_max_axes_delta(ax):
    x_lim = ax.get_xlim()
    y_lim = ax.get_ylim()
    z_lim = ax.get_zlim()
    return max([x_lim[1]-x_lim[0], y_lim[1]-y_lim[0], z_lim[1]-z_lim[0]])


def plot_dof_higlight(ax, model, initial, **options):
    size = get_max_axes_delta(ax)/25 * options.get('plot/symbol_size', 5)

    dof = model.free_dofs[options.get('plot/dof_idx', None)]
    if dof is None:
        return

    node_id, dof_type = dof

    node = model.get_node(node_id)

    dx, dy, dz = 0, 0, 0
    if dof_type == 'u':
        dx = 1
    if dof_type == 'v':
        dy = 1
    if dof_type == 'w':
        dz = 1

    color = 'lightgray' if initial else 'tab:blue'
    if initial:
        x = node.reference_x
        y = node.reference_y
        z = node.reference_z
    else:
        x = node.x
        y = node.y
        z = node.z
    a = get_dof_arrow(x, y, z, dx, dy, dz, size*0.75, color=color)
    ax.add_artist(a)
    # TODO: fix size of spere...
    # a = get_sphere(x, y, z, size/300, color=color)
    # ax.add_artist(a)


def plot_forces(ax, model, initial, **options):
    size = get_max_axes_delta(ax)/25 * options.get('plot/symbol_size', 5)

    for element in model.elements:
        if type(element) == SingleLoad:
            node = element.node
            color = 'lightgray' if initial else 'lightcoral'
            if initial:
                x = node.reference_x
                y = node.reference_y
                z = node.reference_z
            else:
                x = node.x
                y = node.y
                z = node.z
            a = get_force_arrow(x, y, z, element.fu, element.fv, element.fw, size, color=color)

            ax.add_artist(a)


def plot_boundary_conditions(ax, model, initial, **options):
    size = get_max_axes_delta(ax)/100.0 * options.get('plot/symbol_size', 5)

    polygons = list()

    for node in model.nodes:
        for dof_type in ['u', 'v', 'w']:
            if node.dof(dof_type).is_active:
                continue
            if initial:
                polygons.extend(get_tet4_polygons(node.reference_x, node.reference_y, node.reference_z, size, dof_type))
            else:
                polygons.extend(get_tet4_polygons(node.x, node.y, node.z, size, dof_type))

    color = 'lightgray' if initial else 'lightcoral'
    pc = Poly3DCollection(polygons, edgecolor=color, linewidth=0.5, alpha=0.25)
    pc.set_facecolor(color)  # needs to be defined outside otherwhise alpha is not working
    ax.add_collection3d(pc)


def animate_model(fig, ax, models, speed=200, **options):
    bounding_box = get_bounding_box(models)

    def update(step):
        step_model = models[step]

        ax.clear()

        ax.grid()

        plot_bounding_cube(ax, bounding_box)

        ax.set_xlabel('< x >')
        ax.set_ylabel('< y >')
        ax.set_zlabel('< z >')

        title = f'Deformed structure at time step {step}'

        if step_model.name is not None:
            title += f'\n{step}'

        ax.set_title(title)

        plot_model(ax, step_model, 'gray', True, **options)
        plot_model(ax, step_model, 'red', False, **options)

    return anim.FuncAnimation(fig, update, frames=len(models), repeat=True, interval=speed)


def plot_crosshair(ax, x, y, **kwargs):
    lx = ax.axvline(**kwargs)
    lx.set_xdata(x)

    ly = ax.axhline(**kwargs)
    ly.set_ydata(y)


def plot_custom_curve(ax, *args, **kwargs):
    ax.plot(*args, **kwargs)


def plot_bounding_cube(ax, bounding_box, color='w'):
    min_x, max_x, min_y, max_y, min_z, max_z = bounding_box

    xyz_min = np.array([min_x, min_y, min_z])
    xyz_max = np.array([max_x, max_y, max_z])

    max_range = np.array(xyz_max - xyz_min).max()

    center = (xyz_max + xyz_min) / 2

    corners = max_range / 2 * np.mgrid[-1:2:2, -1:2:2, -1:2:2].reshape(3, 8).T + center

    for x, y, z in corners:
        ax.plot([x], [y], [z], color)


def show_animation(model, speed=200, block=True):
    if model.status == ModelStatus.eigenvector:
        return show_eigenvector_animation(model, speed, block)
    else:
        return show_history_animation(model, speed, block)


def show_history_animation(model, speed=200, block=True, size=(10, 10)):
    history = model.get_model_history()

    fig = Figure(figsize=size)
    ax = fig.add_subplot(111, projection='3d')

    ani = animate_model(fig, ax, history, speed=speed)

    return HTML(ani.to_jshtml())


def show_eigenvector_animation(model, speed=200, block=True, size=(10, 10)):
    eigenvector = model
    initial_model = model.get_initial_model()

    models = [initial_model, eigenvector]

    bounding_box = get_bounding_box(models)

    fig = Figure(figsize=size)
    ax = fig.add_subplot(111, projection='3d')

    def update(step):
        step_model = models[step]

        ax.clear()

        ax.grid()

        plot_bounding_cube(ax, bounding_box)

        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')

        ax.title('Eigenvector of Structure \n{}'.format(step_model.name))

        plot_model(ax, step_model, 'gray', True)
        plot_model(ax, step_model, 'red', False)

    ani = anim.FuncAnimation(fig, update, frames=2, repeat=True, interval=speed)

    return HTML(ani.to_jshtml())


def show_deformation_plot(model, step=None, size=(10, 10)):
    bounding_box = get_bounding_box([model])

    fig = Figure(figsize=size)
    ax = fig.add_subplot(111, projection='3d')

    ax.clear()

    ax.grid()
    plot_bounding_cube(ax, bounding_box)

    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')

    plot_model(ax, model, 'gray', True)

    if step is not None:
        model = model.get_model_history()[step]
    else:
        step = len(model.get_model_history())-1

    plot_model(ax, model, 'red', False)

    ax.set_title('Deformed structure at time step {}\n{}'.format(step, model.name))

    fig.show()
